Unit PgRA_LRU;
{$R-,S-,O+}

{ Defines a Paged Generic VirtualArray. MaxSize = 32 MegaBytes.        }

{ Possible Errors and DosExitCodes:

            200 Index Out of Range.
            201 Attempted COPY with Incompatible Data Sizes.
            202 Insufficient Free DiskSpace for COPY operation.
            203 Out of Memory Error encountered during SWAP.
}

INTERFACE

Uses PMan_LRU;

Const
  MaximumSize = 33554432; {32 MegaBytes}

Type
  PathStr = String[79];
  LRUPagedArray = Object

                   ElSize   : Word;
                   NumElems : LongInt;
                   Elems    : PageManager;

                   Procedure Init (NumElements : LongInt; ElementSize : Word;
                                   MaxBuffSize : LongInt; FileName : PathStr);

                   Procedure Load (FileName : String; ElementSize : Word;
                                   MaxBuffSize : LongInt);

                   {NOTE: Performing a LOAD should ONLY be done as a DIRECT}
                   {      substitution for performing an INIT operation}
                   {      Of course, CREATE should be used first.}

                   Procedure Destroy;  {Erases file -- destroys buffers}

                   {By Using LOAD and STORE, it is possible to access a }
                   {file in read-only mode, as used in the TextBrowser  }
                   {object.  -- SEE UNIT P_BROWSE.                      }

                   Procedure Store;

                   {NOTE: Performing a STORE has the same effect as}
                   {      performing a DESTROY, accept the data is}
                   {      saved in the filename given when performing INIT}

           {FileNames May be up to 79 characters long, and may conist
            of Directory and Path information as well as name and extension.
            To Load, BufferedAray MUST be ONLY CREATEd (or DESTROYed)}

                   Procedure Accept (Var El; Index : LongInt; Size : Word);
                   Procedure Retrieve (Var El; Index : LongInt; Size : Word);
                   Procedure Copy (Var From : LRUPagedArray);
                     {NOTE: After a COPY operation, BOTH LRUPagedArrays will}
                     {be using UP TO MemAvail Div 2 for their buffer space}
                   Procedure Swap (I,J : LongInt);

                   Function MaxSize : LongInt;
                   Function ElemSize : Word;
                End;

IMPLEMENTATION

Uses Dos;

Type
  Space = Array[0..0] of Byte;

Procedure Error (Num : Byte);
Begin
  WriteLn;
  Write ('LRUPagedArray ERROR[',Num:1,']: ');
  Case Num of
           0 : WriteLn (' Index Out of Range. ');
           1 : WriteLn (' Attempted COPY with Incompatible Data Sizes. ');
           2 : WriteLn (' Insufficient Free DiskSpace for COPY operation. ');
           3 : WriteLn (' Out of Memory Error encountered during SWAP. ');
          End;
  WriteLn ('**** PROGRAM TERMINATED ****');
  WriteLn;
  Write ('Press <Return> to Continue.... ');
  ReadLn;
  HALT (200+Num)
End;

Procedure LRUPagedArray.Init;
Begin
  ElSize   := ElementSize;
  NumElems := NumElements;
  Elems.Init (NumElements,ElSize,MaxBuffSize,FileName)
End;

Procedure LRUPagedArray.Destroy;
Begin
  ElSize   := 0;
  NumElems := 0;
  Elems.Destroy
End;

Procedure LRUPagedArray.Store;
Begin
  ElSize   := 0;
  NumElems := 0;
  Elems.Release
End;

Procedure LRUPagedArray.Load (FileName : String; ElementSize : Word;
                              MaxBuffSize : LongInt);
Begin
  ElSize := ElementSize;
  NumElems := Elems.Load (ElSize,MaxBuffSize,FileName)
End;

Function LRUPagedArray.MaxSize : LongInt;
Begin
  MaxSize := NumElems
End;

Function LRUPagedArray.ElemSize : Word;
Begin
  ElemSize := ElSize
End;

Procedure LRUPagedArray.Accept (Var El; Index : LongInt; Size : Word);
Begin
  If (Index < 0) or (Index > NumElems-1) Then Error (0);
  Elems.Accept (Index,Size,El)
End;

Procedure LRUPagedArray.Retrieve (Var El; Index : LongInt; Size : Word);
Begin
  If (Index < 0) or (Index > NumElems-1) Then Error (0);
  Elems.Retrieve (Index,Size,El)
End;

Procedure LRUPagedArray.Copy (Var From : LRUPagedArray);
Var
  Temp : PathStr;
  Ch   : Char;
  W    : Word;
  Buff : ^Space;
  NumR : Word;
  NumW : Word;
  Size : Word;

Begin

  If ElSize <> From.ElSize Then Error (1);
  Temp := FExpand (Elems.FName);
  Ch := UpCase(Temp[1]);
  W := Ord(Ch) - Ord('A') + 1;
  If DiskFree(W) < (From.MaxSize * ElSize) Then Error (2);

  Elems.Destroy;

  From.Store; {Flush all buffers, release all memory}
  From.Load (From.Elems.FName,ElSize,MemAvail Div 2);

  GetMem (Buff,2048);

  Assign (Elems.F,Temp);
  ReWrite (Elems.F,1);

  Seek (From.Elems.F,0);

  Repeat
    BlockRead (From.Elems.F,Buff^,2048,NumR);
    BlockWrite (Elems.F,Buff^,NumR,NumW)
  Until (NumR = 0) or (NumW <> NumR);

  Close (Elems.F);

  FreeMem (Buff,2048);

  NumElems := Elems.Load (ElSize,0,Temp)
End;

Procedure LRUPagedArray.Swap (I,J : LongInt);
Var
  T1,T2 : ^Space;
Begin
  GetMem (T1,ElSize);
  GetMem (T2,ElSize);
  If (T1=Nil) or (T2=Nil) Then Error (3);
  Retrieve (T1^,I,ElSize);
  Retrieve (T2^,J,ElSize);
  Accept (T1^,J,ElSize);
  Accept (T2^,I,ElSize);
  FreeMem (T1,ElSize);
  FreeMem (T2,ElSize)
End;

{$F+}
Function HeapErrorTrap (Size : Word) : Integer;
Begin
  HeapErrorTrap := 1  { New and GetMem return Nil if out_of_memory }
End;
{$F-}

BEGIN
  HeapError := @HeapErrorTrap;
END.

