Unit PMan_LRU;
{$R-,S-,O+}

{Project Completion Date : April 7, 1990}

{No modifications to date}

{Uses the Least Recently Used page replacement algorithm}
{Functionally IDENTICAL to PMAN_LFU.}

{Notes on the differences between the two PageManager objects:            }
{Priorities for the LRU algorithm are associated with FRAMES, or In-Memory}
{Pages, as opposed to the LFU algorithm, which keeps a priority for each  }
{page, regardless of its in-memory status. Whenever a page is swapped in, }
{its current frame priority is set equal to the current maximum priority. }
{Calls to ACCEPT and RETRIEVE always increment the priority of the given  }
{frame, and these calls are the only way a page will ever be swapped, so, }
{after the access call which initiated the SWAP, the target frame will    }
{have the highest priority of all frames.  This is opposed to the LFU     }
{concept, which simply increments the page's priority for each access,    }
{thus causing the most frequently demanded pages to stay in memory.       }

{Generalized PageManager Notes:                                            }
{Both PageManagers work with the Page object defined in Unit Pages, and are}
{therefore subject to the same restrictions - in fact they are identical in}
{function, though the LFU version will use a small amount more memory for  }
{its internal overhead, since it must store more priorities (in general).  }
{With that small difference in mind, the following comments apply equally  }
{to either PageManager.                                                    }

{The maximum data size possible (with the current definition of PAGE) is   }
{1536 bytes, or MaxPageSize.  This size was chosen for several reasons:    }
{    (1) Relatively small size to optimize Disk Reading/Writing            }
{    (2) Relatively large size to minimize page-swapping                   }
{    (3) Evenly divisible by 2, 3, 4, 6 and 8 bytes (and more)             }
{        which the designer anticipates will cover the bulk of             }
{        data sizes desired.                                               }
{        NOTE: "Off" sizes present no problem as the Page will             }
{        always allocate to MaxPageSize Div ElementSize.                   }
{        It is merely hoped that an even DOS sector size can               }
{        speed up Disk Reads/Writes by some small factor for               }
{        an even greater overall performance.                              }
{    (4) Data elements larger than 1536 bytes will probably be             }
{        better served by the BufferedArray object, anyway!                }
{        (In fact, I suspect that elements greater than 256 bytes          }
{        would be better off in the BufferedArray!)                        }
{                                                                          }
{As there are several PageManager methods defined which have no Array-like }
{characteristics, the PagedArray objects have been defined as "Having" a   }
{PageManager, rather than Inheriting from PageManager.  This design        }
{decision was motivated by the desire to "shield" the array user from      }
{unnecessary and potentially confusing information.                        }
{                                                                          }
{Unlike the BufferedArray, the PagedArrays permit their buffers to Load and}
{Flush ANYWHERE within the virtual array. This should result in significant}
{performance advantages for the most strenuous exercises, such as sorting, }
{but will probably make no real difference for TOTALLY random operations.  }

{The following PageManager Error messages are defined, and will cause      }
{termination of the program.  For child processes terminated, the errorcode}
{will be available by examining DosExitCode:                               }

{
            100 Unable to Open File "<filename>".
            101 Attempted to create file larger then 32 MegaBytes in size.
            102 Attempted to Initialize with 0 elements.
            103 Unable to LOAD File "<filename>".
            104 Attempted to LOAD with wrong Data Size.
}

INTERFACE

Uses Pages,Crt;

Const
  AbsMaxPageSize = MaxPageSize + SizeOf(Page);
  MaximumSize    = 33554432; {32 MegaBytes}
  Delta          = 0.0000000001;
  MaxPageNum     = MaximumSize Div AbsMaxPageSize;
  MaxFrame       = 426; {MaxMem Div (AbsMaxPageSize + SizeOf(Frame))}

Type
  PathStr    = String[79];
  Frame      = Record
                 PageNum  : LongInt;
                 Priority : Real;
                 Current  : Page;
               End;

  FrameArray = Array[0..0] of Frame; {REMEMBER: Frames require 25 Bytes!}
                                     {FrameArray can hold up to 2620 frames}
                                     {but Actual NumFrames will be <= MaxFrame}
  PageManager = Object

                  Frames     : ^FrameArray;
                  NumFrames  : Word;
                  NumPages   : Word;
                  FName      : PathStr;
                  F          : File;
                  PageLength : Word;  {Number of ELEMENTS in a Page}
                  LowestPty  : Word;  {Index of Frame with Lowest Priority}
                  MaxPty     : Real;  {Current Greatest Priority}
                  CurFrame   : Word;  {Index of current in-use Frame}

                  Procedure Init (NumElements : LongInt; ElementSize : Word;
                                  MaxBuffSize : LongInt; FileName : PathStr);

                  Function Load (ElementSize : Word; MaxBuffSize : LongInt;
                                  FileName : String) : LongInt;

                  Procedure Destroy;  {Erases File}
                  Procedure Release;  {Does Not Erase File}

                  Procedure Retrieve (Index : LongInt; Size : Word; Var El);
                  Procedure Accept (Index : LongInt; Size : Word; Var El);

                  Procedure UpDate_Priorities;
                  Procedure Reset_LowestPty;

                  Function Loaded_At (PageNumber : Word) : Word; {Return Frame}
                  Function Get_Page (Index : LongInt) : Word;  {Return Page}
                  Function Get_Frame (Index : LongInt) : Word; {Return Frame}
                  Procedure Swap_In (PageNumber : Word); {CurFrame = Frame}

                End;

IMPLEMENTATION

Const
  Infinity : Real = 1000000.0; {1 Million should do....}

Type
  Space = Array[0..0] of Byte;

Procedure Error (Num : Byte; Message : String);
Begin
  WriteLn;
  Write ('Page Manager Error [',100+Num,'] : ');
  Case Num of
          0 : WriteLn (' Unable to Open File "',Message,'". ');
          1 : WriteLn (' Attempted to create file larger then 32 MegaBytes in size. ');
          2 : WriteLn (' Attempted to Initialize with 0 elements. ');
          3 : WriteLn (' Unable to LOAD File "',Message,'" ');
          4 : WriteLn (' Attempted to LOAD with wrong Data Size. ');
       End;
  WriteLn;
  WriteLn ('***** PROGRAM TERMINATED *****');
  HALT (100+Num)
End;


Procedure PageManager.Init (NumElements : LongInt; ElementSize : Word;
                            MaxBuffSize : LongInt; FileName : PathStr);
Var
  Temp,I : Word;
  FSize  : LongInt;
  Dummy  : Page;
  FakeEl : ^Space;
  LTemp  : LongInt;
  L      : LongInt;

Begin
  MaxPty := 0.0;
  FName := FileName;
  If NumElements < 1 Then Error (2,'');
  If MaxBuffSize = 0 Then MaxBuffSize := MemAvail;
  If MaxBuffSize > (NumElements * LongInt(ElementSize)) +
                    LongInt(MaxFrame*SizeOf(Frame))
    Then
      MaxBuffSize := (NumElements * LongInt(ElementSize)) +
                      LongInt(MaxFrame*SizeOf(Frame));

  FSize := NumElements * LongInt (ElementSize);
  NumPages := FSize Div Dummy.PageSize(ElementSize);
  If FSize Mod Dummy.PageSize(ElementSize) > 0 Then NumPages := NumPages + 1;

  If (NumPages * LongInt(Dummy.PageSize(ElementSize))) > MaximumSize
    Then
      Error (1,'');

  Temp := System.FileMode;
  System.FileMode := 2;     {Force Open to be for Read/Write}
  {$I-}
  Assign (F,FileName);
  Rewrite (F,1);
  {$I+}
  If IOResult <> 0 Then Error (0,FileName);
  System.FileMode := Temp;  {Put it back the way it was}

  Temp := Dummy.PageSize(ElementSize) + SizeOf(Frame);
  NumFrames := Word (MaxBuffSize Div LongInt(Temp));
  If NumFrames = 0 Then NumFrames := 1;
  If NumFrames > NumPages Then NumFrames := NumPages;
  CurFrame := NumFrames+1;
  PageLength := Dummy.PageSize(ElementSize) Div ElementSize;

  GetMem (Frames,NumFrames*SizeOf(Frame));

  Temp := MaxPageSize Div ElementSize;
  If Temp > NumElements Then
    Begin
      Temp := NumElements;
      PageLength := Temp
    End;
  For I := 0 to NumFrames-1 do  {Initialize all frames}
    With Frames^[I] do
      Begin
        Priority := 0.0;
        PageNum := -1;
        Current.Init (Temp,ElementSize)
      End;

  LowestPty := 0;

  {The ONLY justification for the following block of code is}
  {the expediency of this method for initialization of the  }
  {the disk file.  I apologize to all the purists out there!}
  GetMem (FakeEl,ElementSize);
  For I := 0 to ElementSize-1 do FakeEl^[I] := 0;
  With Frames^[0] do
    Begin
      For L := 0 to Current.HighestEl-1 do
        Current.Accept (L,ElementSize,FakeEl^);
      For L := 0 to NumPages-1 do
        Begin                        {Fill File with Zeroes}
          LTemp := L*(Current.PageSize(ElementSize) Div ElementSize);
          Current.VirtAddr := LTemp;  {VERY,VERY *BAD* ! Don't do this!!!!!}
          While Current.HighestEl >= NumElements do Dec (Current.VirtAddr);
                                     {and ESPECIALLY NEVER do this!!!!!!!!}
          If LTemp < 0 Then {even worse....}
            Begin
              LTemp := 0;
              Current.VirtAddr := 0
            End;
          Current.Accept (LTemp,ElementSize,FakeEl^);
          Current.Flush (F)
        End;
      Current.VirtAddr := -1      {Undo the damage!}
    End;
  FreeMem (FakeEl,ElementSize);
  {End of *BAD* stuff!}

  Temp := 0;
  For I := 0 to NumFrames-1 do {Load the first NUMFRAMES Pages}
  With Frames^[I] do
    Begin
      PageNum := I;
      Current.Load (F,Temp);
      Temp := Temp+PageLength
    End;

End;

Function PageManager.Load;
Var
  Temp,I      : Word;
  FSize       : LongInt;
  Dummy       : Page;
  NumElements : LongInt;

Begin
  MaxPty := 0.0;
  FName := FileName;
  Temp := System.FileMode;
  System.FileMode := 2;     {Force Open to be for Read/Write}
  {$I-}
  Assign (F,FileName);
  Reset (F,1);
  {$I+}
  If IOResult <> 0 Then Error (3,FileName);
  System.FileMode := Temp;  {Put it back the way it was}
  FSize := FileSize (F);

  NumElements := FSize Div ElementSize;
  If FSize Mod NumElements <> 0 Then Error (4,'');

  If MaxBuffSize = 0 Then MaxBuffSize := MemAvail;
  If MaxBuffSize > (NumElements * LongInt(ElementSize)) +
                    LongInt(MaxFrame*SizeOf(Frame))
    Then
      MaxBuffSize := (NumElements * LongInt(ElementSize)) +
                      LongInt(MaxFrame*SizeOf(Frame));

  NumPages := FSize Div Dummy.PageSize(ElementSize);
  If FSize Mod Dummy.PageSize(ElementSize) > 0 Then NumPages := NumPages + 1;

  Temp := Dummy.PageSize(ElementSize) + SizeOf(Frame);
  NumFrames := Word (MaxBuffSize Div LongInt(Temp));
  If NumFrames = 0 Then NumFrames := 1;
  If NumFrames > NumPages Then NumFrames := NumPages;
  CurFrame := NumFrames+1;
  PageLength := Dummy.PageSize(ElementSize) Div ElementSize;

  GetMem (Frames,NumFrames*SizeOf(Frame));

  Temp := MaxPageSize Div ElementSize;
  If Temp > NumElements Then
    Begin
      Temp := NumElements;
      PageLength := Temp
    End;
  For I := 0 to NumFrames-1 do  {Initialize all frames}
    With Frames^[I] do
      Begin
        Priority := 0.0;
        PageNum := -1;
        Current.Init (Temp,ElementSize)
      End;

  LowestPty := 0;

  Temp := 0;
  For I := 0 to NumFrames-1 do {Load the first NUMFRAMES Pages}
  With Frames^[I] do
    Begin
      PageNum := I;
      Current.Load (F,Temp);
      Temp := Temp+PageLength
    End;
  Load := NumElements
End;

Procedure PageManager.Destroy;
Var
  I : Word;
Begin
  For I := 0 to NumFrames-1 do
    With Frames^[I] do
      Begin
        Current.Destroy;
        Priority := 0.0
      End;
  MaxPty := 0.0;
  FreeMem (Frames,NumFrames*SizeOf(Frame));
  Close (F);
  Erase (F)
End;

Procedure PageManager.Release;
Var
  I : Word;
Begin
  For I := 0 to NumFrames-1 do
    With Frames^[I] do
      Begin
        Current.Flush (F);
        Current.Destroy;
        Priority := 0.0
      End;
  MaxPty := 0.0;
  FreeMem (Frames,NumFrames*SizeOf(Frame));
  Close (F);
End;

Procedure PageManager.UpDate_Priorities;
Var
  I    : Word;
  High : Real;
  Low  : Real;
Begin
  High := MaxPty;
  Low  := Infinity;

  For I := 0 to NumFrames-1 do
    With Frames^[I] do
      Begin
        If Priority = Infinity Then Priority := 0.0; {give UP!}
        If Priority < Low Then Low := Priority;
        If Priority > High Then High := Priority;
      End;
  MaxPty := High
End;

Procedure PageManager.Reset_LowestPty;
Var
  I,J : Word;
  Low : Real;
Begin
  Low := Infinity;

  For I := 0 to NumFrames-1 do
    With Frames^[I] do
      If Priority < Low Then
        Begin
          Low := Priority;
          J := I
        End;

  LowestPty := J
End;

Function PageManager.Get_Page (Index : LongInt) : Word;
Begin
  Get_Page := Index Div PageLength
End;

Function PageManager.Loaded_At (PageNumber : Word) : Word;
Var
  I,J  : Word;
  Done : Boolean;
Begin
  J := NumFrames+1;         {Flags NOT Loaded}
  Done := False;
  I := 0;
  While (I <= NumFrames-1) and (Not Done) do
    Begin
      With Frames^[I] do
        If PageNum = PageNumber Then
          Begin
            Done := True;
            J := I
          End;
      Inc (I)
    End;
  Loaded_At := J
End;

Function PageManager.Get_Frame (Index : LongInt) : Word;
Begin
  Get_Frame := Loaded_At (Get_Page (Index))
End;

Procedure PageManager.Swap_In (PageNumber : Word);
Begin
  UpDate_Priorities;
  Reset_LowestPty;
  With Frames^[LowestPty] do
    Begin
      Current.Flush (F);
      Current.Load (F,(PageNumber)*PageLength);
      PageNum := PageNumber;
      Priority := MaxPty
    End;
  CurFrame := LowestPty
End;

Procedure PageManager.Accept (Index : LongInt; Size : Word; Var El);
var
  I : Word;
Begin
  CurFrame := Get_Frame (Index);
  If CurFrame = NumFrames+1 Then Swap_In (Get_Page (Index));

  With Frames^[CurFrame] do
    Begin
      Current.Accept (Index,Size,El);
      Priority := Priority + Delta;
    End
End;

Procedure PageManager.Retrieve (Index : LongInt; Size : Word; Var El);
var
  I : Word;
Begin
  CurFrame := Get_Frame (Index);
  If CurFrame = NumFrames+1 Then Swap_In (Get_Page (Index));

  With Frames^[CurFrame] do
    Begin
      Current.Retrieve (Index,Size,El);
      Priority := Priority + Delta;
    End
End;

{$F+}
Function HeapErrorTrap (Size : Word) : Integer;
Begin
  HeapErrorTrap := 1  { New and GetMem return Nil if out_of_memory }
End;
{$F-}

BEGIN
  HeapError := @HeapErrorTrap
END.
